/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.  
 * 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 * 
 * 3. All advertising materials mentioning features or use of this 
 *    software must display the following acknowledgment: 
 *    "This product includes software developed by the Apache Group 
 *    for use in the Apache HTTP server project (http://www.apache.org/)." 
 * 
 * 4. The names "Apache Server" and "Apache Group" must not be used to 
 *    endorse or promote products derived from this software without 
 *    prior written permission. For written permission, please contact 
 *    apache@apache.org. 
 * 
 * 5. Products derived from this software may not be called "Apache" 
 *    nor may "Apache" appear in their names without prior written 
 *    permission of the Apache Group. 
 * 
 * 6. Redistributions of any form whatsoever must retain the following 
 *    acknowledgment: 
 *    "This product includes software developed by the Apache Group 
 *    for use in the Apache HTTP server project (http://www.apache.org/)." 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY 
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR 
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 * ==================================================================== 
 * 
 * This software consists of voluntary contributions made by many 
 * individuals on behalf of the Apache Group and was originally based 
 * on public domain software written at the National Center for 
 * Supercomputing Applications, University of Illinois, Urbana-Champaign. 
 * For more information on the Apache Group and the Apache HTTP server 
 * project, please see <http://www.apache.org/>. 
 * 
 */ 

#include "httpd.h" 
#include "http_main.h" 
#include "http_log.h" 
#include "http_config.h"	/* for read_config */ 
#include "http_protocol.h"	/* for read_request */ 
#include "http_request.h"	/* for process_request */ 
#include "http_conf_globals.h" 
#include "http_core.h"		/* for get_remote_host */ 
#include "http_vhost.h" 
#include "util_script.h"	/* to force util_script.c linking */ 
#include "util_uri.h" 
#include "fdqueue.h"
#include "acceptlock.h"
#include "http_accept.h"
#include <netinet/tcp.h> 
#include <stdio.h> 

#ifdef USE_SHMGET_SCOREBOARD
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif
 
#include "pthread.h" 

/* Number of cross-process locks we're managing */
static int lock_count;

/* Number of intra-process locks we're managing */
static int intra_lock_count;

/* Intraprocess locking used by other serialization techniques */
static pthread_mutex_t *intra_mutex = NULL;

static void intra_mutex_cleanup(void *foo)
{
    int i;

    for (i = 0; i < intra_lock_count; i++) {
        (void) pthread_mutex_destroy(&intra_mutex[i]);
    }
}

void intra_mutex_init(pool *p, int number_of_locks)
{
    int i;

    intra_lock_count = number_of_locks;
    intra_mutex = (pthread_mutex_t *)ap_palloc(p, intra_lock_count * sizeof(pthread_mutex_t ));
    for (i = 0; i < intra_lock_count; i++) {
        if (pthread_mutex_init(&intra_mutex[i], NULL) != 0) {
            perror("intra_mutex_init");
	    clean_child_exit(APEXIT_CHILDFATAL);
        }
    }
    ap_register_cleanup(p, NULL, intra_mutex_cleanup, ap_null_cleanup);
}

void intra_mutex_on(int locknum)
{
    if ((errno = pthread_mutex_lock(&intra_mutex[locknum])) != 0) {
        ap_log_error(APLOG_MARK, APLOG_EMERG, 
		     (const server_rec *) ap_get_server_conf(),
		     "Error getting intraprocess lock. Exiting!");
    }
}

void intra_mutex_off(int locknum)
{
    if (pthread_mutex_unlock(&intra_mutex[locknum]) != 0) {
        ap_log_error(APLOG_MARK, APLOG_EMERG, 
		     (const server_rec *) ap_get_server_conf(),
		     "Error releasing intraprocess lock. Exiting!");
    }
}

#if defined(USE_FCNTL_SERIALIZED_ACCEPT) || defined(USE_FLOCK_SERIALIZED_ACCEPT)
static void init_lock_fname(pool *p)
{
    /* XXXX possibly bogus cast */
    ap_lock_fname = ap_psprintf(p, "%s.%lu",
        ap_server_root_relative(p, ap_lock_fname), (unsigned long)getpid());
}

static char * expand_lock_fname(pool *p, int i)
{
    return ap_psprintf(p, "%s.%d", ap_lock_fname, i);
}
#endif

#if defined (USE_USLOCK_SERIALIZED_ACCEPT)
/* XXX - Don't know if we need the intraprocess locks here */

#include <ulocks.h>

static ulock_t *uslock = NULL;



void accept_mutex_init(pool *p, int number_of_locks)
{
    ptrdiff_t old;
    usptr_t *us;
    int i;

    lock_count = number_of_locks;
    uslock = (ulock_t *)ap_palloc(p, lock_count * sizeof(ulock_t));

    for (i = 0; i < lock_count; i++) {
    /* default is 8, allocate enough for all the children plus the parent */
        if ((old = usconfig(CONF_INITUSERS, HARD_SERVER_LIMIT + 1)) == -1) {
	    perror("usconfig(CONF_INITUSERS)");
	    exit(-1);
	}
	
	if ((old = usconfig(CONF_LOCKTYPE, US_NODEBUG)) == -1) {
	    perror("usconfig(CONF_LOCKTYPE)");
	    exit(-1);
	}
	if ((old = usconfig(CONF_ARENATYPE, US_SHAREDONLY)) == -1) {
	    perror("usconfig(CONF_ARENATYPE)");
	    exit(-1);
	}
	if ((us = usinit("/dev/zero")) == NULL) {
	    perror("usinit");
	    exit(-1);
	}
	
	if ((uslock[i] = usnewlock(us)) == NULL) {
	    perror("usnewlock");
	    exit(-1);
	}
    }
}

void accept_mutex_on(int locknum)
{
    switch (ussetlock(uslock[locknum])) {
    case 1:
	/* got lock */
	break;
    case 0:
	fprintf(stderr, "didn't get lock\n");
	clean_child_exit(APEXIT_CHILDFATAL);
    case -1:
	perror("ussetlock");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

void accept_mutex_off(int locknum)
{
    if (usunsetlock(uslock[locknum]) == -1) {
	perror("usunsetlock");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

#elif defined (USE_PTHREAD_SERIALIZED_ACCEPT)

/* This code probably only works on Solaris ... but it works really fast
 * on Solaris.  Note that pthread mutexes are *NOT* released when a task
 * dies ... the task has to free it itself.  So we block signals and
 * try to be nice about releasing the mutex.
 */

#include <pthread.h>

static pthread_mutex_t *accept_mutex = (void *)(caddr_t) -1;
static int *have_accept_mutex;
/* The sections surrounded with this ifdef are not needed because of the
 * sigwait thread. They may be turned back on, though, depending on how the
 * port to nonthreaded mode goes */
#ifdef NEED_TO_BLOCK_SIGNALS_AROUND_PTHREAD_CALLS
static sigset_t accept_block_mask;
static sigset_t *accept_previous_masks = NULL;
#endif

void accept_mutex_child_cleanup(void *foo)
{
    int i;
   
    if (accept_mutex != (void *)(caddr_t)-1) {
        for (i = 0; i < lock_count; i++) {
	    if (have_accept_mutex[i]) {
  	        pthread_mutex_unlock(&accept_mutex[i]);
            }
	}
    }
}
void accept_mutex_child_init(pool *p)
{
    ap_register_cleanup(p, NULL, accept_mutex_child_cleanup, ap_null_cleanup);
#ifdef NEED_TO_BLOCK_SIGNALS_AROUND_PTHREAD_CALLS
    accept_previous_masks = ap_palloc(p, lock_count * sizeof(sigset_t));
#endif
}

void accept_mutex_cleanup(void *foo)
{
    int i;
   
    if (accept_mutex != (void *)(caddr_t)-1
        && munmap((caddr_t) accept_mutex,
                  sizeof(pthread_mutex_t) * lock_count)) {
	perror("munmap");
    }
    accept_mutex = (void *)(caddr_t)-1;
}

void accept_mutex_init(pool *p, int number_of_locks)
{
    pthread_mutexattr_t mattr;
    int fd;
    int i;
   
    lock_count = number_of_locks;
    have_accept_mutex = (int *)ap_palloc(p, 
		    lock_count * sizeof(int));

    fd = open("/dev/zero", O_RDWR);
    if (fd == -1) {
        perror("open(/dev/zero)");
        exit(APEXIT_INIT);
    }
    accept_mutex = (pthread_mutex_t *) mmap((caddr_t) 0, 
                           sizeof(pthread_mutex_t) * lock_count, 
    			   PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (accept_mutex == (void *) (caddr_t) - 1) {
        perror("mmap");
        exit(APEXIT_INIT);
    }
    close(fd);
    if ((errno = pthread_mutexattr_init(&mattr))) {
        perror("pthread_mutexattr_init");
        exit(APEXIT_INIT);
    }
    if ((errno = pthread_mutexattr_setpshared(&mattr,
    					  PTHREAD_PROCESS_SHARED))) {
        perror("pthread_mutexattr_setpshared");
        exit(APEXIT_INIT);
    }
    for (i = 0; i < lock_count; i++) {
	if ((errno = pthread_mutex_init(&accept_mutex[i], &mattr))) {
	    perror("pthread_mutex_init");
	    exit(APEXIT_INIT);
	}
    }
    if ((errno = pthread_mutexattr_destroy(&mattr))) {
        perror("pthread_mutexattr_destroy");
        exit(APEXIT_INIT);
    }
#ifdef NEED_TO_BLOCK_SIGNALS_AROUND_PTHREAD_CALLS
    sigfillset(&accept_block_mask);
    sigdelset(&accept_block_mask, SIGHUP);
    sigdelset(&accept_block_mask, SIGTERM);
    sigdelset(&accept_block_mask, SIGWINCH);
#endif
    ap_register_cleanup(p, NULL, accept_mutex_cleanup, ap_null_cleanup);
}

void accept_mutex_on(int locknum)
{
    int err;

#ifdef NEED_TO_BLOCK_SIGNALS_AROUND_PTHREAD_CALLS
    if (pthread_sigmask(SIG_BLOCK, &accept_block_mask,
                        &accept_previous_mask[locknum])) {
	perror("pthread_sigmask(SIG_BLOCK)");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
#endif
    if ((err = pthread_mutex_lock(&accept_mutex[locknum]))) {
	errno = err;
	perror("pthread_mutex_lock");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
    have_accept_mutex[locknum] = 1;
}

void accept_mutex_off(int locknum)
{
    int err;

    if ((err = pthread_mutex_unlock(&accept_mutex[locknum]))) {
	errno = err;
	perror("pthread_mutex_unlock");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
    /* There is a slight race condition right here... if we were to die right
     * now, we'd do another pthread_mutex_unlock.  Now, doing that would let
     * another process into the mutex.  pthread mutexes are designed to be
     * fast, as such they don't have protection for things like testing if the
     * thread owning a mutex is actually unlocking it (or even any way of
     * testing who owns the mutex).
     *
     * If we were to unset have_accept_mutex prior to releasing the mutex
     * then the race could result in the server unable to serve hits.  Doing
     * it this way means that the server can continue, but an additional
     * child might be in the critical section ... at least it's still serving
     * hits.
     */
    have_accept_mutex[locknum] = 0;
#ifdef NEED_TO_BLOCK_SIGNALS_AROUND_PTHREAD_CALLS
    if (pthread_sigmask(SIG_SETMASK, &accept_previous_mask[locknum], NULL)) {
	perror("pthread_sigmask(SIG_SETMASK)");
	clean_child_exit(1);
    }
#endif
}

#elif defined (USE_SYSVSEM_SERIALIZED_ACCEPT)

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#ifdef NEED_UNION_SEMUN
/* it makes no sense, but this isn't defined on solaris */
union semun {
    long val;
    struct semid_ds *buf;
    ushort *array;
};

#endif

static int *sem_id = NULL;
static struct sembuf op_on;
static struct sembuf op_off;

/* We get a random semaphore ... the lame sysv semaphore interface
 * means we have to be sure to clean this up or else we'll leak
 * semaphores.
 */
void accept_mutex_cleanup(void *foo)
{
    union semun ick;
    int i;

    for (i = 0; i < lock_count; i++) {
        if (sem_id[i] < 0)
	  return;
	/* this is ignored anyhow */
	ick.val = 0;
	semctl(sem_id[i], 0, IPC_RMID, ick);
    }
}



void accept_mutex_init(pool *p, int number_of_locks)
{
    union semun ick;
    struct semid_ds buf;
    int i;
   
    lock_count = number_of_locks;
    sem_id = (int *)ap_palloc(p, lock_count * sizeof(int));

    for (i = 0; i < lock_count; i++) {
        /* acquire the semaphore */
        sem_id[i] = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
	if (sem_id[i] < 0) {
	    perror("semget");
	    exit(APEXIT_INIT);
	}
	ick.val = 1;
	if (semctl(sem_id[i], 0, SETVAL, ick) < 0) {
	    perror("semctl(SETVAL)");
	    exit(APEXIT_INIT);
	}
	if (!getuid()) {
	  /* restrict it to use only by the appropriate user_id ... not that this
	   * stops CGIs from acquiring it and dinking around with it.
	   */
	    buf.sem_perm.uid = ap_user_id;
	    buf.sem_perm.gid = ap_group_id;
	    buf.sem_perm.mode = 0600;
	    ick.buf = &buf;
	    if (semctl(sem_id[i], 0, IPC_SET, ick) < 0) {
	      perror("semctl(IPC_SET)");
	      exit(APEXIT_INIT);
	    }
	}
	ap_register_cleanup(p, NULL, accept_mutex_cleanup, ap_null_cleanup);

	/* pre-initialize these */
	op_on.sem_num = 0;
	op_on.sem_op = -1;
	op_on.sem_flg = SEM_UNDO;
	op_off.sem_num = 0;
	op_off.sem_op = 1;
	op_off.sem_flg = SEM_UNDO;
    }
}

void accept_mutex_on(int locknum)
{
    if (semop(sem_id[locknum], &op_on, 1) < 0) {
	perror("accept_mutex_on");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

void accept_mutex_off(int locknum)
{
    if (semop(sem_id[locknum], &op_off, 1) < 0) {
	perror("accept_mutex_off");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

#elif defined(USE_FCNTL_SERIALIZED_ACCEPT)
static struct flock lock_it;
static struct flock unlock_it;

static int *lock_fd = NULL;

/*
 * Initialize mutex lock.
 * Must be safe to call this on a restart.
 */
void accept_mutex_init(pool *p, int number_of_locks)
{
    int i;
    char * lock_fname;

    lock_count = number_of_locks;
    lock_fd = (int *)ap_palloc(p, lock_count * sizeof(int *));
    lock_it.l_whence = SEEK_SET;	/* from current point */
    lock_it.l_start = 0;		/* -"- */
    lock_it.l_len = 0;			/* until end of file */
    lock_it.l_type = F_WRLCK;		/* set exclusive/write lock */
    lock_it.l_pid = 0;			/* pid not actually interesting */
    unlock_it.l_whence = SEEK_SET;	/* from current point */
    unlock_it.l_start = 0;		/* -"- */
    unlock_it.l_len = 0;		/* until end of file */
    unlock_it.l_type = F_UNLCK;		/* set exclusive/write lock */
    unlock_it.l_pid = 0;		/* pid not actually interesting */

    init_lock_fname(p);
    for (i = 0; i < lock_count; i++) {
        lock_fname = expand_lock_fname(p, i);
	lock_fd[i] = ap_popenf(p, lock_fname, 
			       O_CREAT | O_WRONLY | O_EXCL, 0644);
	if (lock_fd[i] == -1) {
	    perror("open");
	    fprintf(stderr, "Cannot open lock file: %s\n", lock_fname);
	    exit(APEXIT_INIT);
	}
	unlink(lock_fname);
    }
}

void accept_mutex_on(int locknum)
{
    int ret;

    while ((ret = fcntl(lock_fd[locknum], F_SETLKW, &lock_it)) < 0 && 
	   errno == EINTR) {
	/* nop */
    }

    if (ret < 0) {
	ap_log_error(APLOG_MARK, APLOG_EMERG, (const server_rec*) ap_get_server_conf(),
		    "fcntl: F_SETLKW: Error getting accept lock, exiting!  "
		    "Perhaps you need to use the LockFile directive to place "
		    "your lock file on a local disk!");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

void accept_mutex_off(int locknum)
{
    int ret;

    while ((ret = fcntl(lock_fd[locknum], F_SETLKW, &unlock_it)) < 0 
	   && errno == EINTR) {
	/* nop */
    }
    if (ret < 0) {
	ap_log_error(APLOG_MARK, APLOG_EMERG, (const server_rec*) ap_get_server_conf(),
		    "fcntl: F_SETLKW: Error freeing accept lock, exiting!  "
		    "Perhaps you need to use the LockFile directive to place "
		    "your lock file on a local disk!");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

#elif defined(USE_FLOCK_SERIALIZED_ACCEPT)

static int *lock_fd = NULL;

void accept_mutex_cleanup(void *foo)
{
    int i;
  
    char * lock_fname;
    for (i = 0; i < lock_count; i++) {
        lock_fname = expand_lock_fname(foo, i);
        unlink(lock_fname);
    }
}

/*
 * Initialize mutex lock.
 * Done by each child at it's birth
 */ 
void accept_mutex_child_init(pool *p)
{
    int i;
  
    for (i = 0; i < lock_count; i++) {
        char *lock_fname = expand_lock_fname(p, i);
    
        lock_fd[i] = ap_popenf(p, lock_fname, O_WRONLY, 0600);
        if (lock_fd[i] == -1) {
	  ap_log_error(APLOG_MARK, APLOG_EMERG, 
		       (const server_rec *)ap_get_server_conf(),
		       "Child cannot open lock file: %s", lock_fname);
	  clean_child_exit(APEXIT_CHILDINIT);
        }
    }
}

/*
 * Initialize mutex lock.
 * Must be safe to call this on a restart.
 */
void accept_mutex_init(pool *p, int number_of_locks)
{
    int i;
    char *lock_fname;
  
    lock_count = number_of_locks;
    lock_fd = (int *)ap_palloc(p, lock_count * sizeof(int *));
    init_lock_fname(p);
    for (i = 0; i < lock_count; i++) {
        lock_fname = expand_lock_fname(p, i);
        unlink(lock_fname);
        lock_fd[i] = ap_popenf(p, lock_fname, 
			       O_CREAT | O_WRONLY | O_EXCL, 0600);
        if (lock_fd[i] == -1) {
	    ap_log_error(APLOG_MARK, APLOG_EMERG, 
			 (const server_rec *) ap_get_server_conf(), 
			 "Parent cannot open lock file: %s", lock_fname);
            exit(APEXIT_INIT);
        }
        ap_register_cleanup(p, p, accept_mutex_cleanup, ap_null_cleanup);
    }
}

void accept_mutex_on(int locknum)
{
    int ret;

    while ((ret = flock(lock_fd[locknum], LOCK_EX)) < 0 && errno == EINTR)
        continue;

    if (ret < 0) {
        ap_log_error(APLOG_MARK, APLOG_EMERG, 
		     (const server_rec *) ap_get_server_conf(),
		     "flock: LOCK_EX: Error getting accept lock. Exiting!");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

void accept_mutex_off(int locknum)
{
    if (flock(lock_fd[locknum], LOCK_UN) < 0) {
        ap_log_error(APLOG_MARK, APLOG_EMERG, 
		     (const server_rec *) ap_get_server_conf(),
		     "flock: LOCK_UN: Error freeing accept lock. Exiting!");
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

#elif defined(USE_OS2SEM_SERIALIZED_ACCEPT)

static HMTX *lock_sem = NULL;

void accept_mutex_cleanup(void *foo)
{ 
    int i;
  
    for (i = 0; i < lock_count; i++) {
        DosReleaseMutexSem(lock_sem[i]);
	DosCloseMutexSem(lock_sem[i]);
    }
}

/*
 * Initialize mutex lock.
 * Done by each child at it's birth
 */
void accept_mutex_child_init(pool *p)
{
    int i;
    
    for(i = 0; i < lock_count; i++) {
        int rc = DosOpenMutexSem(NULL, &lock_sem[locknum]);

	if (rc != 0) {
	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, ap_get_server_conf(),
			 "Child cannot open lock semaphore, rc=%d", rc);
	    clean_child_exit(APEXIT_CHILDINIT);
	}
    }
}
/*
 * Initialize mutex lock.
 * Must be safe to call this on a restart.
 */
void accept_mutex_init(pool *p, int number_of_locks)
{
    int rc;
    int i;

    lock_count = number_of_locks;
    lock_fd = (int *)ap_palloc(p, lock_count * sizeof(int *));

    for (i = 0; i < lock_count; i++) {
        rc = DosCreateMutexSem(NULL, &lock_sem[i], DC_SEM_SHARED, FALSE);

	if (rc != 0) {
	  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, ap_get_server_conf(),
		       "Parent cannot create lock semaphore, rc=%d", rc);
	  exit(APEXIT_INIT);
	}
	
	ap_register_cleanup(p, NULL, accept_mutex_cleanup, ap_null_cleanup);
    }
}

void accept_mutex_on(int locknum)
{
    int rc = DosRequestMutexSem(lock_sem[locknum], SEM_INDEFINITE_WAIT);

    if (rc != 0) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, ap_get_server_conf(),
		    "OS2SEM: Error %d getting accept lock. Exiting!", rc);
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

void accept_mutex_off(int locknum)
{
    int rc = DosReleaseMutexSem(lock_sem[locknum]);
    
    if (rc != 0) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, ap_get_server_conf(),
		    "OS2SEM: Error %d freeing accept lock. Exiting!", rc);
	clean_child_exit(APEXIT_CHILDFATAL);
    }
}

#elif defined(USE_TPF_CORE_SERIALIZED_ACCEPT)

static int tpf_core_held;

static void accept_mutex_cleanup(void *foo)
{
    if(tpf_core_held)
        coruc(RESOURCE_KEY);
}

#define accept_mutex_init(x)

static void accept_mutex_child_init(pool *p)
{
    ap_register_cleanup(p, NULL, accept_mutex_cleanup, ap_null_cleanup);
    tpf_core_held = 0;
}

static void accept_mutex_on(void)
{
    corhc(RESOURCE_KEY);
    tpf_core_held = 1;
    ap_check_signals();
}

static void accept_mutex_off(void)
{
    coruc(RESOURCE_KEY);
    tpf_core_held = 0;
    ap_check_signals();
}

#else
/* Default --- no serialization.  Other methods *could* go here,
 * as #elifs...
 */
#if !defined(MULTITHREAD)
/* Multithreaded systems don't complete between processes for
 * the sockets. */

#endif
#endif

