#include "fdqueue.h"

/* Assumption: queue itself is allocated by the user */
/* Assumption: increment and decrement are atomic on int */

int queue_init(FDQueue *queue, int queue_capacity, pool *a) {
    int i;
    int bounds = queue_capacity + 1;
    pthread_mutex_init(&queue->one_big_mutex, NULL);
    pthread_cond_init(&queue->not_empty, NULL);
    pthread_cond_init(&queue->not_full, NULL);
    queue->head = queue->tail = 0;
    queue->data = ap_palloc(a, bounds * sizeof(FDQueueElement));
    queue->bounds = bounds;
    queue->blanks = 0;
    ap_register_cleanup(a, queue, (void (*)(void *))queue_destroy, ap_null_cleanup);
    for (i=0; i < bounds; ++i)
        queue->data[i].fd = -1;
    return FD_QUEUE_SUCCESS;
}

void *queue_destroy(FDQueue *queue) {
    /* Ignore errors here, we can't do anything about them anyway */
    pthread_cond_destroy(&queue->not_empty);
    pthread_cond_destroy(&queue->not_full);
    pthread_mutex_destroy(&queue->one_big_mutex);
    return FD_QUEUE_SUCCESS;
}

int queue_push(FDQueue *queue, int fd, struct sockaddr *addr) {
    if (pthread_mutex_lock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    queue->data[queue->tail].fd = fd;
    queue->data[queue->tail].addr = *addr;
    queue->tail = (queue->tail + 1) % queue->bounds;
    queue->blanks--;
    pthread_cond_signal(&queue->not_empty);
    if (queue->head == (queue->tail + 1) % queue->bounds) {
        pthread_cond_wait(&queue->not_full, &queue->one_big_mutex);
    }
    if (pthread_mutex_unlock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    return FD_QUEUE_SUCCESS;
}

int queue_pop(FDQueue *queue, struct sockaddr *addr, int block_if_empty) {
    int fd;
    if (pthread_mutex_lock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    if (queue->blanks > 0) {
        pthread_cond_signal(&queue->not_full);
    }
    if (queue->head == queue->tail) {
        if (block_if_empty)
            pthread_cond_wait(&queue->not_empty, &queue->one_big_mutex);
        else {
            pthread_mutex_unlock(&queue->one_big_mutex);
            return FD_QUEUE_FAILURE;
        }
    } 
    
    fd = queue->data[queue->head].fd;
    *addr = queue->data[queue->head].addr;
    queue->data[queue->head].fd = -1;
    if (fd != -1) {
        queue->head = (queue->head + 1) % queue->bounds;
    }
    if (pthread_mutex_unlock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    return fd;
}

int queue_size(FDQueue *queue) {
    return ((queue->tail - queue->head + queue->bounds) % queue->bounds);
}

int queue_full(FDQueue *queue) {
    return(queue->blanks <= 0);
}

int block_on_queue(FDQueue *queue) {
    if (pthread_mutex_lock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    pthread_cond_wait(&queue->not_full, &queue->one_big_mutex);
    if (pthread_mutex_unlock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    return FD_QUEUE_SUCCESS;
}

int increase_blanks(FDQueue *queue) {
    if (pthread_mutex_lock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    queue->blanks++;
    if (pthread_mutex_unlock(&queue->one_big_mutex) != 0) {
        return FD_QUEUE_FAILURE;
    }
    return FD_QUEUE_SUCCESS;
}
