/* ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.  
 * 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in 
 *    the documentation and/or other materials provided with the 
 *    distribution. 
 * 
 * 3. All advertising materials mentioning features or use of this 
 *    software must display the following acknowledgment: 
 *    "This product includes software developed by the Apache Group 
 *    for use in the Apache HTTP server project (http://www.apache.org/)." 
 * 
 * 4. The names "Apache Server" and "Apache Group" must not be used to 
 *    endorse or promote products derived from this software without 
 *    prior written permission. For written permission, please contact 
 *    apache@apache.org. 
 * 
 * 5. Products derived from this software may not be called "Apache" 
 *    nor may "Apache" appear in their names without prior written 
 *    permission of the Apache Group. 
 * 
 * 6. Redistributions of any form whatsoever must retain the following 
 *    acknowledgment: 
 *    "This product includes software developed by the Apache Group 
 *    for use in the Apache HTTP server project (http://www.apache.org/)." 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY 
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR 
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 * ==================================================================== 
 * 
 * This software consists of voluntary contributions made by many 
 * individuals on behalf of the Apache Group and was originally based 
 * on public domain software written at the National Center for 
 * Supercomputing Applications, University of Illinois, Urbana-Champaign. 
 * For more information on the Apache Group and the Apache HTTP server 
 * project, please see <http://www.apache.org/>. 
 * 
 */ 
 
#include "httpd.h" 
#include "http_log.h"
#include "http_main.h"
#include "http_conf_globals.h" 
#include "acceptlock.h"
#include "fdqueue.h"
#include "scoreboard.h"
#include "http_accept.h"
#include <poll.h>
#include <netinet/tcp.h>

/* Indicates that all acceptor threads are dead after SIGWINCH and the worker
 * threads can now exit */
static int workers_may_exit = 0;
static int requests_this_child;
static int num_listenfds;

#if defined (USE_ACCEPT_QUEUE)

#ifdef SINGLE_LISTEN_UNSERIALIZED_ACCEPT
/* Each thread only listens to one socket in this model, so the starvation
 * problem described in manual/misc/perf-tuning.html can't occur here */
#define SAFE_ACCEPT(stmt) 
#else
#define SAFE_ACCEPT(stmt) do {stmt;} while(0)
#endif

/* The queue of sockets we've accepted */
static FDQueue csd_queue;

static void * accept_thread(void * dummy)
{
    proc_info * ti = dummy;
    int my_pid = ti->pid;
    int my_tid = ti->tid;
    int sd = ti->sd;
    int csd = 0;
    int srv, ret;
    struct sockaddr sa_client;
    size_t len = sizeof(struct sockaddr);
    struct pollfd listenfds[2];
    char pipe_read_char;

    free(ti);

    /*
     * Let's setup the socket options on the master socket.  These
     * will be inherited by any dup'ed sockets for us.  No reason to do 
     * this oce for each request.

    sock_disable_nagle(sd);
    */

    listenfds[0].fd = ap_pipe_of_death[0];
    listenfds[0].events = POLLIN;
    listenfds[0].revents = 0;
    listenfds[1].fd = sd;
    listenfds[1].events = POLLIN;
    listenfds[1].revents = 0;

    while ((ap_max_requests_per_child != 0 && requests_this_child > 0) ||
           (ap_max_requests_per_child == 0)) { 
        if (workers_may_exit)
            break;
        (void) ap_update_child_status(my_pid, my_tid, SERVER_ACCEPTING, 
				    (request_rec *) NULL);
	/* lock around the accept if necessary */
	SAFE_ACCEPT(intra_mutex_on(my_tid - ap_threads_per_child));
	if (workers_may_exit) {
	    SAFE_ACCEPT(intra_mutex_off(my_tid - ap_threads_per_child));
	    break;
	}
	SAFE_ACCEPT(accept_mutex_on(my_tid - ap_threads_per_child));
        if (queue_full(&csd_queue)) {
            SAFE_ACCEPT(accept_mutex_off(my_tid - ap_threads_per_child));
            SAFE_ACCEPT(intra_mutex_off(my_tid - ap_threads_per_child));
            block_on_queue(&csd_queue);
            csd = -1;
        }
        else {
            for (;;) {
                srv = poll(listenfds, 2, -1);
                if (listenfds[0].revents & POLLIN) {
                    /* A process has gotten a signal on the shutdown pipe.
                     * Check if we're the lucky process to die. */
                    ret = read(listenfds[0].fd, &pipe_read_char, 1);
                    if (ret == -1 && errno == EAGAIN) {
                        /* It lost the lottery. It must continue to suffer
                         * through a life of servitude */
                        continue;
                    }
                    else {
                        /* It won the lottery (or something else is very
                         * wrong).  Embrace death with open arms. */
                        workers_may_exit = 1;
                        pthread_cond_broadcast(&(csd_queue.not_empty));
                        csd = -1;
                        break;
                    }
                } else {
                    csd = ap_accept(sd, &sa_client, &len);
                    requests_this_child--;
		    break;
                }
            }
            SAFE_ACCEPT(accept_mutex_off(my_tid - ap_threads_per_child));
            SAFE_ACCEPT(intra_mutex_off(my_tid - ap_threads_per_child));
	}
        (void) ap_update_child_status(my_pid, my_tid, SERVER_QUEUEING, 
				      (request_rec *) NULL);
	if (csd >= 0) {
	    if (queue_push(&csd_queue, csd, &sa_client) != 0) {
	        ap_log_error(APLOG_MARK, APLOG_ERR, ap_get_server_conf(),
			 "queue_push:  couldn't put new connection on"
                         "queue");
	    }
        } 
    }

    ap_update_child_status(my_pid, my_tid, SERVER_DEAD, (request_rec *) NULL);
    return NULL;
}

void accept_parent_init(pool *pconf, int listener_count)
{
    SAFE_ACCEPT(accept_mutex_init(pconf, listener_count));
    ap_acceptors_per_child = num_listenfds = listener_count;
}

/*
 * Description:
 * Do any setup or initialization required before worker threads begin 
 * accepting connections
 * Design questions:
 * 1. what sort of arguments need to be passed on this call?
 * 2. should the queue logic be simplified?
 * 3. Pass in server_conf?
 * 4. Simply access the globals (yech...)
 */
void accept_child_init(pool* pchild, 
                 int worker_threads_per_child)
{
    int queue_capacity = worker_threads_per_child + ap_acceptors_per_child;
    SAFE_ACCEPT(intra_mutex_init(pchild, ap_acceptors_per_child));
    SAFE_ACCEPT(accept_mutex_child_init(pchild));
    requests_this_child = ap_max_requests_per_child;
    queue_init(&csd_queue, queue_capacity, pchild);
}


void start_accepting_connections(int my_child_num)
{
    proc_info *my_info;
    pthread_t thread;
    /*
     * Yech, ap_listeners is a global declared in http_conf_globals.h 
     * and defined in http_main.c 
     */
    listen_rec *lr = ap_listeners;


    /* Warning: Subtle code ahead
     * 1. This code assumes the thing pointed to by ap_listeners is a NULL terminated
     * linked list. 
     *
     * 2. The scoreboard size is ap_threads_per_child + ap_acceptors_per_child
     * The acceptor entries are at the end of the scoreboard, so index into the
     * scoreboard beginning with ap_threads_per_child
     * Perhaps a better approach is to create a new scoreboard accessor method to
     * fetch the index of the next free entry? And put in some bounds checking
     * to prevent running off the end of the scoreboard? (ie, if next_free_entry() 
     * returns NULL and we find ourselves needing another entry, we have a logic
     * bug that needs to be fixed...
     * Perhaps we should create a scoreboard object with methods, etc.?
     * ap_threads_per_child is a GLOBAL...
     */
    int i = ap_threads_per_child;
    if (lr) {
        while (lr != NULL) {

	    (void) ap_update_child_status(my_child_num, i, SERVER_STARTING, 
					  (request_rec *) NULL);

	    my_info = NULL;

	    my_info = (proc_info *) malloc(sizeof(proc_info));
	    my_info->pid = my_child_num;
	    my_info->tid = i++;
	    my_info->sd = lr->fd;


	    if (pthread_create(&thread, NULL, accept_thread, my_info)) {
	        ap_log_error(APLOG_MARK, APLOG_ALERT, (const server_rec*) ap_get_server_conf(),
			     "pthread_create: unable to create acceptor thread");
		/* 
		 * We failed to create a thread.  Update the scoreboard,  
		 * or it will say SERVER_STARTING forever.
		 */
		exit(1);
	    } else {
	      /* We let each thread update it's own scoreboard entry.  This is done
	       * because it let's us deal with tid better.
	       */
	    }
	    lr = lr->next;
	}
    } else {
      /* no listening sockets????  Kill the server please.  */
      exit(0);
    }
}

int get_connection(struct sockaddr *sa_client)
{
    int csd = -1;
    int block_if_empty = 1;

    increase_blanks(&csd_queue);

    while (csd == -1) {
        if (workers_may_exit)
            block_if_empty = 0;

        csd = queue_pop(&csd_queue, sa_client, block_if_empty);

        if (workers_may_exit)
            break;
    }

    return csd;

}

#elif defined(USE_MULTI_ACCEPT)

#ifdef SINGLE_LISTEN_UNSERIALIZED_ACCEPT
#define SAFE_ACCEPT(stmt) do {if(ap_listeners->next != NULL) {stmt;}} while(0)
#else
#define SAFE_ACCEPT(stmt) do {stmt;} while(0)
#endif

/*
 * USE_MULTI_ACCEPT
 * Worker threads do the accept and process the request. 
 */
static listen_rec *head_listener;
static struct pollfd *listenfds;

void accept_parent_init(pool *pconf, int listener_count)
{
    SAFE_ACCEPT(accept_mutex_init(pconf, 1));
    num_listenfds = listener_count;
    ap_acceptors_per_child = 0;
}

void accept_child_init(pool* pchild, 
                 int worker_threads_per_child)
{
    int i;
    listen_rec *lr;

    SAFE_ACCEPT(intra_mutex_init(pchild, 1));
    SAFE_ACCEPT(accept_mutex_child_init(pchild));
    requests_this_child = ap_max_requests_per_child;
    head_listener = ap_listeners;

    listenfds = ap_palloc(pchild, sizeof(struct pollfd) * (num_listenfds + 1));
    listenfds[0].fd = ap_pipe_of_death[0];
    listenfds[0].events = POLLIN;
    listenfds[0].revents = 0;
    for (lr = ap_listeners, i = 1; i <= num_listenfds; lr = lr->next, ++i) {
	lr->index = i;
	listenfds[i].fd = lr->fd;
	listenfds[i].events = POLLIN; /* should we add POLLPRI ?*/
	listenfds[i].revents = 0;
    }
}

void start_accepting_connections(int my_child_num) 
{
}
int get_connection(struct sockaddr *sa_client) 
{
    int csd = -1;
    int sd;
    int srv;
    int ret;
    listen_rec *lr;
    char pipe_read_char;

    size_t len = sizeof(struct sockaddr);

    while ((ap_max_requests_per_child != 0 && requests_this_child > 0) ||
           (ap_max_requests_per_child == 0)) {

	SAFE_ACCEPT(intra_mutex_on(0));
	if (workers_may_exit) {
	    SAFE_ACCEPT(intra_mutex_off(0));
	    return -1;
	}
	SAFE_ACCEPT(accept_mutex_on(0));

        for (;;) {

            if (workers_may_exit)
                break;

            srv = poll(listenfds, num_listenfds + 1, -1);
            if (listenfds[0].revents & POLLIN) {
                /* A process has gotten a signal on the shutdown pipe.
                 * Check if we're the lucky process to die. */
                ret = read(listenfds[0].fd, &pipe_read_char, 1);
                if (ret == -1 && errno == EAGAIN) {
                    /* It lost the lottery. It must continue to suffer through
                     * a life of servitude */
                    continue;
                }
                else {
                    /* It won the lottery (or something else is very wrong).
                     * Embrace death with open arms. */
                    break;
                }
            }
            if (workers_may_exit)
                break;
            if (srv < 0) {
                /* XXX - hmmmm, poll can return EINTR. - mvsk */
                if (errno == EINTR) {
                    continue;
                }
                SAFE_ACCEPT(accept_mutex_off(0));
                SAFE_ACCEPT(intra_mutex_off(0));
                /* Single Unix documents select as returning errnos
                 * EBADF, EINTR, and EINVAL... and in none of those
                 * cases does it make sense to continue.  In fact
                 * on Linux 2.0.x we seem to end up with EFAULT
                 * occasionally, and we'd loop forever due to it.
                 */
                ap_log_error(APLOG_MARK, APLOG_ERR, (const server_rec*) ap_get_server_conf(), "select: (listen)");
                clean_child_exit(1);
            }
            
            if (srv == 0) {
                /* Is srv == 0 a valid return? */
                continue;
            }

	    /* This conditional is used because the single listen case is the
	     * only one where the accept might not be serialized. In that
	     * case, multiple threads mucking around with the head_listener
	     * pointer will be harmful */
            if (num_listenfds == 1) {
	        /* only one socket, just pretend we did the other stuff */
	        sd = ap_listeners->fd;
            }
            else {
                /* find a listener */
                /* Loop or NULL terminated list? That is the question. Be
                 * consistent across all the accept techniques */
                lr = head_listener;
                do {
                    /* XXX: should we check for PR_POLL_ERR ?? */
                    if (listenfds[lr->index].revents & POLLIN) {
                        /* advance to the next listener for next loop */
                        head_listener = lr->next;
                        /* hack to handle listenfds being NULL terminated list 
                         * rather than a loop 
                         */
                        if (head_listener == NULL) {
                            head_listener = ap_listeners;
                        }
                        goto got_lr;
                    }
                    lr = lr->next;
                    if (lr == NULL) {
    			lr = ap_listeners;
                    }
                } while (lr != head_listener);
                
                /* if we don't find anything then just start again */
                fprintf(stderr,"poll returned but we got nothing!\n");	
                head_listener = ap_listeners;	
                continue;
    
            got_lr:
                sd = lr->fd;
            }
            csd = ap_accept(sd, sa_client, &len);   
            requests_this_child--;

            if (csd >= 0)
                break;		/* We have a socket ready for reading */
            /* XXX: we need to deal with error conditions here */
        }

        SAFE_ACCEPT(accept_mutex_off(0));
        SAFE_ACCEPT(intra_mutex_off(0));

	if (csd >= 0) {
            return csd;
        } else{
            break;
        }
    }
    workers_may_exit = 1;
    return -1;
}

#endif

